
/**
 * @brief Blinky Sample Application main file.
 *
 * This file contains the source code for a sample server application using the LED Button service.
 */

#include <stdint.h>
#include <string.h>

#include "ble.h"
#include "nrf.h"
#include "fds.h"
#include "boards.h"
#include "ble_bas.h"
#include "ble_err.h"
#include "ble_hci.h"
#include "ble_lbs.h"
#include "nrf_sdh.h"
#include "ble_racp.h"
#include "app_error.h"
#include "app_timer.h"
#include "app_button.h"
#include "bsp_btn_ble.h"
#include "ble_advdata.h"
#include "nrf_sdh_soc.h"
#include "nrf_sdh_ble.h"
#include "nrf_fstorage.h"
#include "nrf_ble_gatt.h"
#include "peer_manager.h"
#include "nrf_ble_gatt.h"
#include "nordic_common.h"
#include "nrf_drv_saadc.h"
#include "ble_conn_state.h"
#include "ble_srv_common.h"
#include "ble_advertising.h"
#include "ble_conn_params.h"
#include "nrf_fstorage_sd.h"


#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

/* Custom service libraries */
#include "ble_tpls.h"
#include "ble_custom_dns.h"

/* Macro Definetions ---------------------------------------------------------------------------------------- */
/* ------------------------------ ADC Settings ------------------------------------------------------------ */
#define ADC_REF_VOLTAGE_IN_MILLIVOLTS   600                                     /**< Reference voltage (in milli volts) used by ADC while doing conversion. */
#define ADC_PRE_SCALING_COMPENSATION    6                                       /**< The ADC is configured to use VDD with 1/3 prescaling as input. And hence the result of conversion is to be multiplied by 3 to get the actual value of the battery voltage.*/
#define DIODE_FWD_VOLT_DROP_MILLIVOLTS  270                                     /**< Typical forward voltage drop of the diode . */
#define ADC_RES_10BIT                   1024                                    /**< Maximum digital value for 10-bit ADC conversion. */

/* ------------------------------ Battery Level Settings --------------------------------------------------- */
#define BATTERY_LEVEL_MEAS_INTERVAL     APP_TIMER_TICKS(2000)                   /**< Battery level measurement interval (ticks). */
#define MIN_BATTERY_LEVEL               81                                      /**< Minimum battery level as returned by the simulated measurement function. */
#define MAX_BATTERY_LEVEL               100                                     /**< Maximum battery level as returned by the simulated measurement function. */
#define BATTERY_LEVEL_INCREMENT         1                                       /**< Value by which the battery level is incremented/decremented for each call to the simulated measurement function. */

/* ------------------------------ Device Information -------------------------------------------------------- */
#define GROUP_ID						0x00									/**< Group id of device */
#define PASS_KEY						"123456"								/**< Connection password */
#define GROUP_NAME						"GrpName"								/**< Group name of device. */
#define DEVICE_NAME                     "ButtonApp"                            	/**< Name of device. Will be included in the advertising data. */
#define APP_FEATURE_NOT_SUPPORTED       BLE_GATT_STATUS_ATTERR_APP_BEGIN + 2    /**< Reply when unsupported features are requested. */

/* ------------------------------ Flash Settings --------------------------------------------------- */
#define APP_FLASH_START_ADDRESS	0x3E000
#define APP_FLASH_END_ADDRESS	0x3FFFF

/* ------------------------------ Interval & Timeout ------------------------------------------------------- */
#define APP_ADV_INTERVAL                64                                      /**< The advertising interval (in units of 0.625 ms; this value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED   /**< The advertising time-out (in units of seconds). When set to 0, we will never time out. */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(100, UNIT_1_25_MS)        /**< Minimum acceptable connection interval (0.5 seconds). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(200, UNIT_1_25_MS)        /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                       /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(4000, UNIT_10_MS)         /**< Connection supervisory time-out (4 seconds). */

#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000)	                /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (15 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000)                  /**< Time between each call to sd_ble_gap_conn_param_update after the first call (5 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                       /**< Number of attempts before giving up the connection parameter negotiation. */

#define SECURITY_REQUEST_DELAY          APP_TIMER_TICKS(400)                        /**< Delay after connection until Security Request is sent, if necessary (ticks). */

/* ------------------------------ Leds & Buttons ---------------------------------------------------------- */
#define LEDBUTTON_BUTTON_0              BSP_BUTTON_0                            /**< Button that will trigger the notification event with the LED Button Service */
#define LEDBUTTON_BUTTON_1              BSP_BUTTON_1                            /**< Button that will trigger the notification event with the LED Button Service */
#define LEDBUTTON_BUTTON_2              BSP_BUTTON_2                            /**< Button that will trigger the notification event with the LED Button Service */
#define LEDBUTTON_BUTTON_3              BSP_BUTTON_3                            /**< Button that will trigger the notification event with the LED Button Service */

#define ADVERTISING_LED                 BSP_BOARD_LED_0                         /**< Is on when device is advertising. */
#define CONNECTED_LED                   BSP_BOARD_LED_1                         /**< Is on when device has connected. */
#define LEDBUTTON_LED                   BSP_BOARD_LED_2                         /**< LED to be toggled with the help of the LED Button Service. */

#define BUTTON_DETECTION_DELAY          APP_TIMER_TICKS(50)                     /**< Delay from a GPIOTE event until a button is reported as pushed (in number of timer ticks). */

/* ------------------------------ Tx Power ---------------------------------------------------------------- */
#define TX_POWER_LEVEL                  (-8)                                    /**< TX Power Level value. This will be set both in the TX Power service, in the advertising data, and also used to set the radio transmit power. */

/* ------------------------------ Observer Settings -------------------------------------------------------- */
#define APP_BLE_OBSERVER_PRIO           3                                       /**< Application's BLE observer priority. You shouldn't need to modify this value. */
#define APP_BLE_CONN_CFG_TAG            1                                       /**< A tag identifying the SoftDevice BLE configuration. */

/* ------------------------------ Security Parameters ----------------------------------------------------- */
#define SEC_PARAM_BOND                  1                                           /**< Perform bonding. */
#define SEC_PARAM_MITM                  1                                           /**< Man In The Middle protection required (applicable when display module is detected). */
#define SEC_PARAM_LESC                  0                                           /**< LE Secure Connections not enabled. */
#define SEC_PARAM_KEYPRESS              0                                           /**< Keypress notifications not enabled. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_DISPLAY_ONLY                /**< Display I/O capabilities. */
#define SEC_PARAM_OOB                   0                                           /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                           /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                          /**< Maximum encryption key size. */
#define PASSKEY_TXT                     "Passkey:"                                  /**< Message to be displayed together with the pass-key. */
#define PASSKEY_TXT_LENGTH              8                                           /**< Length of message to be displayed together with the pass-key. */
#define PASSKEY_LENGTH                  6                                           /**< Length of pass-key received by the stack for display. */

#define DEAD_BEEF                       0xDEADBEEF                              /**< Value used as error code on stack dump, can be used to identify stack location on stack unwind. */

/* Private Types & Macro-------------------------------------------------------------------------------------  */

/**@brief Macro to convert the result of ADC conversion in millivolts.
 *
 * @param[in]  ADC_VALUE   ADC result.
 *
 * @retval     Result converted to millivolts.
 */
#define ADC_RESULT_IN_MILLI_VOLTS(ADC_VALUE)\
        ((((ADC_VALUE) * ADC_REF_VOLTAGE_IN_MILLIVOLTS) / ADC_RES_10BIT) * ADC_PRE_SCALING_COMPENSATION)
		
/* size must not exceed ( ( x * 4 ) - 2 ) because of size of struct. Otherwise compiler will padding it to the 4x. */
#define MAX_NAME_LENGTH		16
#define BOOT_SIGNATURE		0xEEAA
typedef struct
{
	/**< */
	uint16_t write_identifier;
	/**< */
	int8_t tx_power;
	uint8_t grp_id;
	
	/**< advertisement parameters */
	uint16_t adv_timeout;
	uint16_t adv_interval;
	
	/**< gap parameters */
	uint16_t min_conn_int;
	uint16_t max_conn_int;
	
	/**< */
	
	char grp_name[ MAX_NAME_LENGTH ];
	char dev_name[ MAX_NAME_LENGTH ];
}config_t;

/* Function prototypes -------------------------------------------------------------------------------------- */

static void log_init(void);
static void gatt_init(void);
static void leds_init(void);
static void timers_init(void);
static void buttons_init(void);
static void delete_bonds(void);
static void power_manage(void);
static void adc_configure(void);
static void services_init(void);
static void ble_stack_init(void);
static void gap_params_init(void);
static void sleep_mode_enter(void);
static void conn_params_init(void);
static void advertising_init(void);
static void advertising_start(void);
static void wait_for_flash_ready(void);
static void application_timers_start(void);

/* Handler prototypes -------------------------------------------------------------------------------------  */
static void pm_evt_handler(pm_evt_t const * p_evt);
static void sec_req_timeout_handler(void * p_context);
static void conn_params_error_handler(uint32_t nrf_error);
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt);
static void on_bas_evt(ble_bas_t * p_bas, ble_bas_evt_t * p_evt);
static void fstorage_evt_handler( nrf_fstorage_evt_t * p_fds_evt);
static void ble_tpls_handler( ble_tpls_t *ble_tlps, int8_t tx_power );
static void ble_dns_grpid_handler( ble_dns_t * p_dns, uint8_t grp_id );
static void button_event_handler(uint8_t pin_no, uint8_t button_action);
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context);
static void led_write_handler(uint16_t conn_handle, ble_lbs_t * p_lbs, uint8_t led_state);
static void ble_dns_name_handler( ble_dns_t *p_dns, const uint8_t *name, uint16_t length );
static void ble_dns_grpname_handler( ble_dns_t * p_dns, const uint8_t * grp_name, uint16_t length );

/* Global Variables -------------------------------------------------------------------------------------  */
BLE_BAS_DEF(m_bas);                                                    	 		/**< Structure used to identify the battery service. */
BLE_LBS_DEF(m_lbs);                                                             /**< LED Button Service instance. */
//BLE_DNS_DEF(m_dns);															/**< Device Name Service instance */
NRF_BLE_GATT_DEF(m_gatt);                                                       /**< GATT module instance. */
APP_TIMER_DEF(m_battery_timer_id);                                      		/**< Battery timer. */
APP_TIMER_DEF(m_sec_req_timer_id);                                              /**< Security Request timer. */
BLE_ADVERTISING_DEF(m_advertising);                                             /**< Advertising module instance. */
NRF_FSTORAGE_DEF(nrf_fstorage_t fstorage) =
{
    /* Set a handler for fstorage events. */
    .evt_handler = fstorage_evt_handler,

    /* These below are the boundaries of the flash space assigned to this instance of fstorage.
     * You must set these manually, even at runtime, before nrf_fstorage_init() is called.
     * The function nrf5_flash_end_addr_get() can be used to retrieve the last address on the
     * last page of flash available to write data. */
    .start_addr = APP_FLASH_START_ADDRESS,
    .end_addr   = APP_FLASH_END_ADDRESS,
};

ble_dns_t				m_dns;
ble_tpls_t				m_tpls;													/**< Tx Power Level Service instance */
ble_advdata_t 			advdata;
ble_tpls_init_t 		tpls_init;
ble_gap_conn_sec_mode_t sec_mode;
/* Private Variables -------------------------------------------------------------------------------------  */

static bool pending_write = false;
static uint8_t  pass_key[]		= PASS_KEY;
static uint8_t	buttons			= 0;
static uint16_t	m_conn_handle 	= BLE_CONN_HANDLE_INVALID;                  	/**< Handle of the current connection. */

static nrf_saadc_value_t adc_buf[2];

static pm_peer_id_t m_peer_to_be_deleted = 	PM_PEER_ID_INVALID;
static ble_opt_t ble_opt = { .gap_opt.passkey.p_passkey = pass_key };

static config_t config = 
{
	.write_identifier = BOOT_SIGNATURE,
		
	.grp_id = GROUP_ID,
	.tx_power = TX_POWER_LEVEL,
	
	.adv_timeout = APP_ADV_TIMEOUT_IN_SECONDS,
	.adv_interval = APP_ADV_INTERVAL,
	
	.min_conn_int = 100,
	.max_conn_int = 200,
	
	.grp_name = { GROUP_NAME },
	.dev_name = { DEVICE_NAME }
};
//

/* Callback functions --------------------------------------------------------------------------------------- */

/**@brief Function for assert macro callback.
 *
 * @details This function will be called in case of an assert in the SoftDevice.
 *
 * @warning This handler is an example only and does not fit a final product. You need to analyze
 *          how your product is supposed to react in case of Assert.
 * @warning On assert from the SoftDevice, the system can only recover on reset.
 *
 * @param[in] line_num    Line number of the failing ASSERT call.
 * @param[in] p_file_name File name of the failing ASSERT call.
 */
void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
    app_error_handler(DEAD_BEEF, line_num, p_file_name);
}
//

/* Handler Functions --------------------------------------------------------------------------------------- */
/**@brief Function for handling a Connection Parameters error.
 *
 * @param[in] nrf_error  Error code containing information about what went wrong.
 */
static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}
//

static void fstorage_evt_handler( nrf_fstorage_evt_t * p_evt )
{
    if (p_evt->result != NRF_SUCCESS)
    {
        NRF_LOG_INFO("--> Event received: ERROR while executing an fstorage operation.");
        return;
    }

    switch (p_evt->id)
    {
        case NRF_FSTORAGE_EVT_WRITE_RESULT:
        {
            NRF_LOG_INFO("--> Event received: wrote %d bytes at address 0x%x.",
                         p_evt->len, p_evt->addr);
        } break;

        case NRF_FSTORAGE_EVT_ERASE_RESULT:
        {
            NRF_LOG_INFO("--> Event received: erased %d page from address 0x%x.",
                         p_evt->len, p_evt->addr);
        } break;

        default:
            break;
    }
}
//

/**@brief Function for handling the Battery Service events.
 *
 * @details This function will be called for all Battery Service events which are passed to the
 |          application.
 *
 * @param[in] p_bas  Battery Service structure.
 * @param[in] p_evt  Event received from the Battery Service.
 */
static void on_bas_evt(ble_bas_t * p_bas, ble_bas_evt_t * p_evt)
{
    ret_code_t err_code;

    switch (p_evt->evt_type)
    {
        case BLE_BAS_EVT_NOTIFICATION_ENABLED:
            // Start battery timer
            err_code = app_timer_start(m_battery_timer_id, BATTERY_LEVEL_MEAS_INTERVAL, NULL);
            APP_ERROR_CHECK(err_code);
            break; // BLE_BAS_EVT_NOTIFICATION_ENABLED

        case BLE_BAS_EVT_NOTIFICATION_DISABLED:
            err_code = app_timer_stop(m_battery_timer_id);
            APP_ERROR_CHECK(err_code);
            break; // BLE_BAS_EVT_NOTIFICATION_DISABLED

        default:
            // No implementation needed.
            break;
    }
}
//

static void ble_dns_grpid_handler( ble_dns_t * p_dns, uint8_t grp_id )
{
	NRF_LOG_INFO( "Change group id: %d", grp_id );
	
	config.grp_id = grp_id;
	pending_write = true;
}
//

/**@brief Function for handling write events to the LED characteristic.
 *
 * @param[in] p_lbs     Instance of LED Button Service to which the write applies.
 * @param[in] led_state Written/desired state of the LED.
 */
static void led_write_handler(uint16_t conn_handle, ble_lbs_t * p_lbs, uint8_t led_state)
{
    if (led_state)
    {
        bsp_board_led_on(LEDBUTTON_LED);
        NRF_LOG_INFO("Received LED ON!");
    }
    else
    {
        bsp_board_led_off(LEDBUTTON_LED);
        NRF_LOG_INFO("Received LED OFF!");
    }
}
//

static void ble_dns_name_handler( ble_dns_t *p_dns, const uint8_t *dev_name, uint16_t length )
{
	uint32_t err_code;
	
	char * p_str = ( char * )dev_name;
	p_str[ length ] = '\0';
	
	NRF_LOG_INFO( "Change device name: %s", nrf_log_push( p_str ) );
    
	memset( config.dev_name, 0, MAX_NAME_LENGTH );
	memcpy( config.dev_name, p_str, length );
	
	pending_write = true;
	
	err_code = sd_ble_gap_device_name_set( &sec_mode, dev_name, length );
    APP_ERROR_CHECK(err_code);

    err_code = ble_advdata_set(&advdata, NULL);
    APP_ERROR_CHECK(err_code);
	
//	advertising_start();
}
//

static void ble_dns_grpname_handler( ble_dns_t * p_dns, const uint8_t * grp_name, uint16_t length )
{
	char * p_str = ( char * )grp_name;
	// terminate string
	p_str[ length ] = '\0';
	// print debug
	NRF_LOG_INFO( "Change group name: %s", p_str );
    
	// clear group name buffer
	memset( config.grp_name, 0, MAX_NAME_LENGTH );
	// copy string to the group name buffer
	strcpy( config.grp_name, p_str );
	
	pending_write = true;
}
//

/* Event handler functions ---------------------------------------------------------------------------------- */

/**@brief Function for handling advertising events.
 *
 * @param[in] ble_adv_evt  Advertising event.
 */
static void on_adv_evt(ble_adv_evt_t ble_adv_evt)
{
    ret_code_t err_code;

    switch (ble_adv_evt)
    {
        case BLE_ADV_EVT_FAST:
            NRF_LOG_INFO("Fast advertising");
//            err_code = bsp_indication_set(BSP_INDICATE_ADVERTISING);
            APP_ERROR_CHECK(err_code);
            break; // BLE_ADV_EVT_FAST

        case BLE_ADV_EVT_IDLE:
            sleep_mode_enter();
            break; // BLE_ADV_EVT_IDLE

        default:
            break;
    }
}
//

/**@brief Function for handling Peer Manager events.
 *
 * @param[in] p_evt  Peer Manager event.
 */
static void pm_evt_handler(pm_evt_t const * p_evt)
{
    ret_code_t err_code;

    switch (p_evt->evt_id)
    {
        case PM_EVT_BONDED_PEER_CONNECTED:
        {
            NRF_LOG_INFO("Connected to a previously bonded device.");
            // Start Security Request timer.
            err_code = app_timer_start(m_sec_req_timer_id, SECURITY_REQUEST_DELAY, NULL);
            APP_ERROR_CHECK(err_code);
        } break;

        case PM_EVT_CONN_SEC_SUCCEEDED:
        {
            pm_conn_sec_status_t conn_sec_status;

            // Check if the link is authenticated (meaning at least MITM).
            err_code = pm_conn_sec_status_get(p_evt->conn_handle, &conn_sec_status);
            APP_ERROR_CHECK(err_code);

            if (conn_sec_status.mitm_protected)
            {
                NRF_LOG_INFO("Link secured. Role: %d. conn_handle: %d, Procedure: %d",
                             ble_conn_state_role(p_evt->conn_handle),
                             p_evt->conn_handle,
                             p_evt->params.conn_sec_succeeded.procedure);
            }
            else
            {
                // The peer did not use MITM, disconnect.
                NRF_LOG_INFO("Collector did not use MITM, disconnecting");
                err_code = pm_peer_id_get(m_conn_handle, &m_peer_to_be_deleted);
                APP_ERROR_CHECK(err_code);
                err_code = sd_ble_gap_disconnect(m_conn_handle,
                                                 BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
                APP_ERROR_CHECK(err_code);
            }
        } break;

        case PM_EVT_CONN_SEC_FAILED:
        {
            NRF_LOG_INFO("Failed to secure connection. Disconnecting.");
            err_code = sd_ble_gap_disconnect(m_conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            if (err_code != NRF_ERROR_INVALID_STATE)
            {
                APP_ERROR_CHECK(err_code);
            }
            m_conn_handle = BLE_CONN_HANDLE_INVALID;
        } break;

        case PM_EVT_CONN_SEC_CONFIG_REQ:
        {
            // Reject pairing request from an already bonded peer.
            pm_conn_sec_config_t conn_sec_config = {.allow_repairing = false};
            pm_conn_sec_config_reply(p_evt->conn_handle, &conn_sec_config);
        } break;

        case PM_EVT_STORAGE_FULL:
        {
            // Run garbage collection on the flash.
            err_code = fds_gc();
            if (err_code == FDS_ERR_BUSY || err_code == FDS_ERR_NO_SPACE_IN_QUEUES)
            {
                // Retry.
            }
            else
            {
                APP_ERROR_CHECK(err_code);
            }
        } break;

        case PM_EVT_PEERS_DELETE_SUCCEEDED:
        {
            NRF_LOG_DEBUG("PM_EVT_PEERS_DELETE_SUCCEEDED [ advertising_start ]");
            advertising_start(/*false*/);
        } break;

        case PM_EVT_LOCAL_DB_CACHE_APPLY_FAILED:
        {
            // The local database has likely changed, send service changed indications.
            pm_local_database_has_changed();
        } break;

        case PM_EVT_PEER_DATA_UPDATE_FAILED:
        {
            // Assert.
            APP_ERROR_CHECK(p_evt->params.peer_data_update_failed.error);
        } break;

        case PM_EVT_PEER_DELETE_FAILED:
        {
            // Assert.
            APP_ERROR_CHECK(p_evt->params.peer_delete_failed.error);
        } break;

        case PM_EVT_PEERS_DELETE_FAILED:
        {
            // Assert.
            APP_ERROR_CHECK(p_evt->params.peers_delete_failed_evt.error);
        } break;

        case PM_EVT_ERROR_UNEXPECTED:
        {
            // Assert.
            APP_ERROR_CHECK(p_evt->params.error_unexpected.error);
        } break;

        case PM_EVT_CONN_SEC_START:
        case PM_EVT_PEER_DATA_UPDATE_SUCCEEDED:
        case PM_EVT_PEER_DELETE_SUCCEEDED:
        case PM_EVT_LOCAL_DB_CACHE_APPLIED:
        case PM_EVT_SERVICE_CHANGED_IND_SENT:
        case PM_EVT_SERVICE_CHANGED_IND_CONFIRMED:
        default:
            break;
    }
}
//

/**@brief Function for handling the Security Request timer timeout.
 *
 * @details This function will be called each time the Security Request timer expires.
 *
 * @param[in] p_context  Pointer used for passing some arbitrary information (context) from the
 *                       app_start_timer() call to the timeout handler.
 */
static void sec_req_timeout_handler(void * p_context)
{
    ret_code_t err_code;

    if (m_conn_handle != BLE_CONN_HANDLE_INVALID)
    {
        // Initiate bonding.
        NRF_LOG_DEBUG("Start encryption");
        err_code = pm_conn_secure(m_conn_handle, false);
        if (err_code != NRF_ERROR_INVALID_STATE)
        {
            APP_ERROR_CHECK(err_code);
        }
    }
}
//

/**@brief Function for handling the ADC interrupt.
 *
 * @details  This function will fetch the conversion result from the ADC, convert the value into
 *           percentage and send it to peer.
 */
void saadc_event_handler(nrf_drv_saadc_evt_t const * p_event)
{
    if (p_event->type == NRF_DRV_SAADC_EVT_DONE)
    {
        nrf_saadc_value_t adc_result;
        uint16_t          batt_lvl_in_milli_volts;
        uint8_t           percentage_batt_lvl;
        uint32_t          err_code;

        adc_result = p_event->data.done.p_buffer[0];

        err_code = nrf_drv_saadc_buffer_convert(p_event->data.done.p_buffer, 1);
        APP_ERROR_CHECK(err_code);

        batt_lvl_in_milli_volts = ADC_RESULT_IN_MILLI_VOLTS(adc_result) +
                                  DIODE_FWD_VOLT_DROP_MILLIVOLTS;
        percentage_batt_lvl = battery_level_in_percent(batt_lvl_in_milli_volts);

        err_code = ble_bas_battery_level_update(&m_bas, percentage_batt_lvl);
        if (
            (err_code != NRF_SUCCESS)
            &&
            (err_code != NRF_ERROR_INVALID_STATE)
            &&
            (err_code != NRF_ERROR_RESOURCES)
            &&
            (err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING)
           )
        {
            APP_ERROR_HANDLER(err_code);
        }
    }
}
//

/**@brief Function for handling the Connection Parameters Module.
 *
 * @details This function will be called for all events in the Connection Parameters Module that
 *          are passed to the application.
 *
 * @note All this function does is to disconnect. This could have been done by simply
 *       setting the disconnect_on_fail config parameter, but instead we use the event
 *       handler mechanism to demonstrate its use.
 *
 * @param[in] p_evt  Event received from the Connection Parameters Module.
 */
static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
    ret_code_t err_code;

    if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED)
    {
        err_code = sd_ble_gap_disconnect(m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);
    }
}
//

/**@brief Set the radio's transmit power.
 *
 * @param[in] tx_power Radio transmit power in dBm (accepted values are -40, -20, -16, -12, -8, -4, 0, 3, and 4 dBm).
 *
 * @retval ::NRF_SUCCESS Successfully changed the transmit power.
 * @retval ::NRF_ERROR_INVALID_PARAM Invalid parameter(s) supplied.
 */
void ble_tpls_handler( ble_tpls_t *ble_tlps, int8_t tx_power )
{
	uint32_t err_code;
	
	if( tx_power < -39 )
		tx_power = -40;
	else if ( tx_power > -40 && tx_power < -29 )
		tx_power = -40;
	else if	( tx_power > -30 && tx_power < -15 )
		tx_power = -20;
	else if	( tx_power > -16 && tx_power < -14 )
		tx_power = -16;
	else if	( tx_power > -15 && tx_power < -10 )
		tx_power = -12;
	else if	( tx_power > -11 && tx_power < -5 )
		tx_power = -8;
	else if	( tx_power > -6 && tx_power < -2 )
		tx_power = -4;
	else if	( tx_power > -3 && tx_power < 1 )
		tx_power = 0;
	else if	( tx_power > 1 && tx_power < 4 )
		tx_power = 3;
	else if	( tx_power > 3 )
		tx_power = 4;
	
	NRF_LOG_INFO( "Requested Tx Power: %d", tx_power );
	
	err_code = sd_ble_gap_tx_power_set( tx_power );
	APP_ERROR_CHECK( err_code );
	
	config.tx_power = tx_power;
	pending_write = true;
}
//

/**@brief Function for handling the Battery measurement timer timeout.
 *
 * @details This function will be called each time the battery level measurement timer expires.
 *          This function will start the ADC.
 *
 * @param[in] p_context   Pointer used for passing some arbitrary information (context) from the
 *                        app_start_timer() call to the timeout handler.
 */
static void battery_level_meas_timeout_handler(void * p_context)
{
    UNUSED_PARAMETER(p_context);

    ret_code_t err_code;
    err_code = nrf_drv_saadc_sample();
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for handling events from the button handler module.
 *
 * @param[in] pin_no        The pin that the event applies to.
 * @param[in] button_action The button action (press/release).
 */
static void button_event_handler(uint8_t pin_no, uint8_t button_action)
{
    ret_code_t err_code;

    switch (pin_no)
    {
        case LEDBUTTON_BUTTON_0:
		{
			if( button_action )
				buttons |= 0x00000001;
			else
				buttons &= 0x0000000E;
			break;
		}
        case LEDBUTTON_BUTTON_1:
		{
			if( button_action )
				buttons |= 0x00000002;
			else
				buttons &= 0x0000000D;
			break;
		}
        case LEDBUTTON_BUTTON_2:
		{
			if( button_action )
				buttons |= 0x00000004;
			else
				buttons &= 0x0000000B;
			break;
		}
        case LEDBUTTON_BUTTON_3:
		{
			if( button_action )
				buttons |= 0x00000008;
			else
				buttons &= 0x00000007;
			break;
		}

        default:
			NRF_LOG_INFO("Unknown button pin number %d.", pin_no);
            APP_ERROR_HANDLER(pin_no);
            break;
    }
	
	NRF_LOG_INFO("Send button %d state change.", pin_no);
	
	err_code = ble_lbs_on_button_change(m_conn_handle, &m_lbs, buttons);
	
	if (err_code != NRF_SUCCESS &&
		err_code != BLE_ERROR_INVALID_CONN_HANDLE &&
		err_code != NRF_ERROR_INVALID_STATE &&
		err_code != BLE_ERROR_GATTS_SYS_ATTR_MISSING)
	{
		APP_ERROR_CHECK(err_code);
	}
}
//

/**@brief Function for handling BLE events.
 *
 * @param[in]   p_ble_evt   Bluetooth stack event.
 * @param[in]   p_context   Unused.
 */
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
    ret_code_t err_code;

    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
        {
            NRF_LOG_INFO("Connected");
            m_peer_to_be_deleted = PM_PEER_ID_INVALID;
//            err_code = bsp_indication_set(BSP_INDICATE_CONNECTED);
            APP_ERROR_CHECK(err_code);
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
            // Start Security Request timer.
            err_code = app_timer_start(m_sec_req_timer_id, SECURITY_REQUEST_DELAY, NULL);
            APP_ERROR_CHECK(err_code);
            bsp_board_led_on(CONNECTED_LED);
            bsp_board_led_off(ADVERTISING_LED);
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;

            err_code = app_button_enable();
            APP_ERROR_CHECK(err_code);
            break;
        }

        case BLE_GAP_EVT_DISCONNECTED:
		{
            NRF_LOG_INFO("Disconnected");
            m_conn_handle = BLE_CONN_HANDLE_INVALID;
            
			// Check if the last connected peer had not used MITM, if so, delete its bond information.
            if (m_peer_to_be_deleted != PM_PEER_ID_INVALID)
            {
                err_code = pm_peer_delete(m_peer_to_be_deleted);
                APP_ERROR_CHECK(err_code);
                NRF_LOG_DEBUG("Collector's bond deleted");
                m_peer_to_be_deleted = PM_PEER_ID_INVALID;
            }
            break;
		}
        case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
		{
            NRF_LOG_DEBUG("BLE_GAP_EVT_SEC_PARAMS_REQUEST");
            break;
		}
#ifndef S140
        case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
        {
            NRF_LOG_DEBUG("PHY update request.");
            ble_gap_phys_t const phys =
            {
                .rx_phys = BLE_GAP_PHY_AUTO,
                .tx_phys = BLE_GAP_PHY_AUTO,
            };
            err_code = sd_ble_gap_phy_update(p_ble_evt->evt.gap_evt.conn_handle, &phys);
            APP_ERROR_CHECK(err_code);
        } break;
#endif

        case BLE_GATTC_EVT_TIMEOUT:
		{
            // Disconnect on GATT Client timeout event.
            NRF_LOG_DEBUG("GATT Client Timeout.");
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gattc_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break;
		}
        case BLE_GATTS_EVT_TIMEOUT:
		{
            // Disconnect on GATT Server timeout event.
            NRF_LOG_DEBUG("GATT Server Timeout.");
            err_code = sd_ble_gap_disconnect(p_ble_evt->evt.gatts_evt.conn_handle,
                                             BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
            APP_ERROR_CHECK(err_code);
            break;
		}
        case BLE_GAP_EVT_PASSKEY_DISPLAY:
        {
            char passkey[PASSKEY_LENGTH + 1];
            memcpy(passkey, p_ble_evt->evt.gap_evt.params.passkey_display.passkey, PASSKEY_LENGTH);
            passkey[PASSKEY_LENGTH] = 0;
            // Don't send delayed Security Request if security procedure is already in progress.
            err_code = app_timer_stop(m_sec_req_timer_id);
            APP_ERROR_CHECK(err_code);

            NRF_LOG_INFO("Passkey: %s", nrf_log_push(passkey));
        } break;

        case BLE_EVT_USER_MEM_REQUEST:
		{
            err_code = sd_ble_user_mem_reply(p_ble_evt->evt.gattc_evt.conn_handle, NULL);
            APP_ERROR_CHECK(err_code);
            break;
		}
        case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
        {
            ble_gatts_evt_rw_authorize_request_t  req;
            ble_gatts_rw_authorize_reply_params_t auth_reply;

            req = p_ble_evt->evt.gatts_evt.params.authorize_request;

            if (req.type != BLE_GATTS_AUTHORIZE_TYPE_INVALID)
            {
                if ((req.request.write.op == BLE_GATTS_OP_PREP_WRITE_REQ)     ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_NOW) ||
                    (req.request.write.op == BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL))
                {
                    if (req.type == BLE_GATTS_AUTHORIZE_TYPE_WRITE)
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_WRITE;
                    }
                    else
                    {
                        auth_reply.type = BLE_GATTS_AUTHORIZE_TYPE_READ;
                    }
                    auth_reply.params.write.gatt_status = APP_FEATURE_NOT_SUPPORTED;
                    err_code = sd_ble_gatts_rw_authorize_reply(p_ble_evt->evt.gatts_evt.conn_handle,
                                                               &auth_reply);
                    APP_ERROR_CHECK(err_code);
                }
            }
        } break; // BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST

        default:
            // No implementation needed.
            break;
    }
}
//

/* Initialize functions ------------------------------------------------------------------------------------- */
/**@brief Function for configuring ADC to do battery level conversion. */
static void adc_configure(void)
{
    ret_code_t err_code = nrf_drv_saadc_init(NULL, saadc_event_handler);
    APP_ERROR_CHECK(err_code);

    nrf_saadc_channel_config_t config =
        NRF_DRV_SAADC_DEFAULT_CHANNEL_CONFIG_SE(NRF_SAADC_INPUT_VDD);
    err_code = nrf_drv_saadc_channel_init(0, &config);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_saadc_buffer_convert(&adc_buf[0], 1);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_saadc_buffer_convert(&adc_buf[1], 1);
    APP_ERROR_CHECK(err_code);
}
//

static void log_init(void)
{
    ret_code_t err_code = NRF_LOG_INIT(NULL);
    APP_ERROR_CHECK(err_code);

    NRF_LOG_DEFAULT_BACKENDS_INIT();
}
//

/**@brief Function for initializing the GATT module.
 */
static void gatt_init(void)
{
    ret_code_t err_code = nrf_ble_gatt_init(&m_gatt, NULL);
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for the LEDs initialization.
 *
 * @details Initializes all LEDs used by the application.
 */
static void leds_init(void)
{
    bsp_board_leds_init();
}
//

/**@brief Function for the Timer initialization.
 *
 * @details Initializes the timer module.
 */
static void timers_init(void)
{
    // Initialize timer module, making it use the scheduler
    ret_code_t err_code = app_timer_init();
    APP_ERROR_CHECK(err_code);
	
    // Create timers.
    err_code = app_timer_create(&m_battery_timer_id,
                                APP_TIMER_MODE_REPEATED,
                                battery_level_meas_timeout_handler);
    APP_ERROR_CHECK(err_code);
	
    // Create Security Request timer.
    err_code = app_timer_create(&m_sec_req_timer_id,
                                APP_TIMER_MODE_SINGLE_SHOT,
                                sec_req_timeout_handler);
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for initializing the button handler module.
 */
static void buttons_init(void)
{
    ret_code_t err_code;

    //The array must be static because a pointer to it will be saved in the button handler module.
    static app_button_cfg_t buttons[] =
    {
        {LEDBUTTON_BUTTON_0, false, BUTTON_PULL, button_event_handler},
        {LEDBUTTON_BUTTON_1, false, BUTTON_PULL, button_event_handler},
        {LEDBUTTON_BUTTON_2, false, BUTTON_PULL, button_event_handler},
        {LEDBUTTON_BUTTON_3, false, BUTTON_PULL, button_event_handler}
    };

    err_code = app_button_init(buttons, sizeof(buttons) / sizeof(buttons[0]),
                               BUTTON_DETECTION_DELAY);
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for initializing services that will be used by the application.
 */
static void services_init(void)
{
    ret_code_t     	err_code;
    ble_lbs_init_t 	init;
    ble_bas_init_t 	bas_init;
	ble_dns_init_t 	dns_init;

	/**< BATTERY SERVICE INITIALIZE */
	
	// Initialize Battery Service Service
    memset( &bas_init, 0, sizeof( bas_init ) );
	
    /* Here the security level for the Battery Service notification permission. */
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&bas_init.battery_level_char_attr_md.cccd_write_perm);
    /* Here the security level for the Battery Service read permission. */
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&bas_init.battery_level_char_attr_md.read_perm);
    /* Here the security level for the Battery Service no write permission. */
    BLE_GAP_CONN_SEC_MODE_SET_NO_ACCESS(&bas_init.battery_level_char_attr_md.write_perm);
    /* Here the security level for the Battery Service notification read permission. */
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&bas_init.battery_level_report_read_perm);
	
    bas_init.evt_handler          = on_bas_evt;
    bas_init.support_notification = true;
    bas_init.p_report_ref         = NULL;
    bas_init.initial_batt_level   = 100;

    err_code = ble_bas_init(&m_bas, &bas_init);
    APP_ERROR_CHECK(err_code);

	/**< DEVICE NAME SERVICE INITIALIZE */
	
	// Initialize Device Name Service
    memset( &dns_init, 0, sizeof( dns_init ) );
	
	/* Here the security level for Device Name Service read permission */
	BLE_GAP_CONN_SEC_MODE_SET_OPEN( &dns_init.dns_attr_md.read_perm );
	/* Here the security level for Device Name Service write permission */
	BLE_GAP_CONN_SEC_MODE_SET_OPEN( &dns_init.dns_attr_md.write_perm );
	
	dns_init.group_id = config.grp_id;
	
	ble_srv_ascii_to_utf8( &dns_init.group_name_str, config.grp_name );
	ble_srv_ascii_to_utf8( &dns_init.device_name_str, config.dev_name );
	
	m_dns.device_name_handler 	= ble_dns_name_handler;
	m_dns.group_id_handler 		= ble_dns_grpid_handler;
	m_dns.group_name_handler 	= ble_dns_grpname_handler;
	
	err_code = ble_dns_init( &m_dns, &dns_init );
	APP_ERROR_CHECK( err_code );

    // Register a handler for Device Name Service events.
    NRF_SDH_BLE_OBSERVER(m_dns_observer, APP_BLE_OBSERVER_PRIO, ble_dns_on_ble_evt, &m_dns);
	
	/**< LED BUTTON SERVICE INITIALIZE */
	// Initialize LED Button Service
    init.led_write_handler = led_write_handler;

    err_code = ble_lbs_init(&m_lbs, &init);
    APP_ERROR_CHECK(err_code);
	
	/**< TX POWER LEVEL SERVICE INITIALIZE */
	
	// Initialize TX Power Level Service
    memset( &tpls_init, 0, sizeof( tpls_init ) );

    /* Here the security level for the Tx Power Level Service read permission. */
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&tpls_init.tpls_att_md.read_perm );
    /* Here the security level for the Tx Power Level Service write permission. */
	BLE_GAP_CONN_SEC_MODE_SET_OPEN(&tpls_init.tpls_att_md.write_perm );
	
	tpls_init.tx_power_level = config.tx_power;
	m_tpls.tx_power_level_handler = ble_tpls_handler;
	
	err_code = ble_tpls_init( &m_tpls, &tpls_init );
    APP_ERROR_CHECK(err_code);
	
    // Register a handler for Tx Power Level Service events.
    NRF_SDH_BLE_OBSERVER(m_tpls_observer, APP_BLE_OBSERVER_PRIO, ble_tpls_on_ble_evt, &m_tpls);
}
//

/**@brief Function for initializing the BLE stack.
 *
 * @details Initializes the SoftDevice and the BLE event interrupt.
 */
static void ble_stack_init(void)
{
    ret_code_t err_code;

    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    // Configure the BLE stack using the default settings.
    // Fetch the start address of the application RAM.
    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    // Enable BLE stack.
    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    // Register a handler for BLE events.
    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}
//

/**@brief Function for the GAP initialization.
 *
 * @details This function sets up all the necessary GAP (Generic Access Profile) parameters of the
 *          device including the device name, appearance, and the preferred connection parameters.
 */
static void gap_params_init(void)
{
    ret_code_t              err_code;
    ble_gap_conn_params_t   gap_conn_params;
//    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                          (const uint8_t *)config.dev_name,
                                          strlen(config.dev_name));
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MSEC_TO_UNITS( config.min_conn_int, UNIT_1_25_MS );
    gap_conn_params.max_conn_interval = MSEC_TO_UNITS( config.max_conn_int, UNIT_1_25_MS );
    gap_conn_params.slave_latency     = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout  = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for initializing the Advertising functionality.
 *
 * @details Encodes the required advertising data and passes it to the stack.
 *          Also builds a structure to be passed to the stack when starting advertising.
 */
static void advertising_init(void)
{
	ret_code_t    			err_code;
	ble_advertising_init_t 	init;

    ble_uuid_t adv_uuids[] = 
	{
		{ LBS_UUID_SERVICE,				BLE_UUID_TYPE_BLE },
		{ BLE_UUID_BATTERY_SERVICE,		BLE_UUID_TYPE_BLE },
		{ BLE_UUID_TPC_SERVICE_UUID,	BLE_UUID_TYPE_BLE }

	};

	memset( &init, 0, sizeof( init ) );
	
    init.advdata.name_type               = BLE_ADVDATA_FULL_NAME;
    init.advdata.include_appearance      = true;
    init.advdata.flags                   = BLE_GAP_ADV_FLAGS_LE_ONLY_LIMITED_DISC_MODE;
    init.advdata.uuids_complete.uuid_cnt = sizeof( adv_uuids ) / sizeof( adv_uuids[0] );
    init.advdata.uuids_complete.p_uuids  = adv_uuids;

    init.config.ble_adv_fast_enabled  = true;
    init.config.ble_adv_fast_interval = MSEC_TO_UNITS( config.adv_interval, UNIT_1_25_MS );//APP_ADV_INTERVAL;
    init.config.ble_adv_fast_timeout  = MSEC_TO_UNITS( 200, UNIT_1_25_MS );//APP_ADV_TIMEOUT_IN_SECONDS;

    init.evt_handler = on_adv_evt;

    err_code = ble_advertising_init(&m_advertising, &init);
    APP_ERROR_CHECK(err_code);

    ble_advertising_conn_cfg_tag_set(&m_advertising, APP_BLE_CONN_CFG_TAG);
}
//

/**@brief Function for initializing the Connection Parameters module.
 */
static void conn_params_init(void)
{
    ret_code_t             err_code;
    ble_conn_params_init_t cp_init;

    memset(&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = NULL;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay  = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count   = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = false;
    cp_init.evt_handler                    = on_conn_params_evt;
    cp_init.error_handler                  = conn_params_error_handler;

    err_code = ble_conn_params_init(&cp_init);
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for the Peer Manager initialization.
 */
static void peer_manager_init(void)
{
    ble_gap_sec_params_t sec_param;
    ret_code_t           err_code;

    err_code = pm_init();
    APP_ERROR_CHECK(err_code);

    memset(&sec_param, 0, sizeof(ble_gap_sec_params_t));

    // Security parameters to be used for all security procedures.
    sec_param.bond           = SEC_PARAM_BOND;
    sec_param.mitm           = SEC_PARAM_MITM;
    sec_param.lesc           = SEC_PARAM_LESC;
    sec_param.keypress       = SEC_PARAM_KEYPRESS;
    sec_param.io_caps        = SEC_PARAM_IO_CAPABILITIES;
    sec_param.oob            = SEC_PARAM_OOB;
    sec_param.min_key_size   = SEC_PARAM_MIN_KEY_SIZE;
    sec_param.max_key_size   = SEC_PARAM_MAX_KEY_SIZE;
    sec_param.kdist_own.enc  = 1;
    sec_param.kdist_own.id   = 1;
    sec_param.kdist_peer.enc = 1;
    sec_param.kdist_peer.id  = 1;

    err_code = pm_sec_params_set(&sec_param);
    APP_ERROR_CHECK(err_code);

    err_code = pm_register(pm_evt_handler);
    APP_ERROR_CHECK(err_code);
}
//

/* Other functions ------------------------------------------------------------------------------------- */
/**@brief Clear bond information from persistent storage.
 */
static void delete_bonds(void)
{
    ret_code_t err_code;

    NRF_LOG_INFO("Erase bonds!");

    err_code = pm_peers_delete();
    APP_ERROR_CHECK(err_code);
}
//

static void config_init(void)
{
	uint32_t err_code;
	
	// Initialize fstorage
	err_code = nrf_fstorage_init( &fstorage, &nrf_fstorage_sd, NULL );
	APP_ERROR_CHECK( err_code );
	wait_for_flash_ready();
	
	config_t temp_config;
	memset( &temp_config, 0, sizeof( config_t ) );
	
	err_code = nrf_fstorage_read( &fstorage, APP_FLASH_START_ADDRESS, &temp_config, sizeof( config_t ) );
	APP_ERROR_CHECK( err_code );
	wait_for_flash_ready();
	
	// if first boot
	if( temp_config.write_identifier != BOOT_SIGNATURE )
	{
		config.write_identifier = BOOT_SIGNATURE;
		NRF_LOG_INFO( "First time boot" );
		err_code = nrf_fstorage_write( &fstorage, APP_FLASH_START_ADDRESS, &config, sizeof( config_t ), NULL );
		APP_ERROR_CHECK( err_code );
		
		wait_for_flash_ready();
	}
	else
		NRF_LOG_INFO( "Configured before!" );
	
	err_code = nrf_fstorage_read( &fstorage, APP_FLASH_START_ADDRESS, &config, sizeof( config_t ) );
	APP_ERROR_CHECK( err_code );
	wait_for_flash_ready();
}
//

/**@brief Function for the Power Manager.
 */
static void power_manage(void)
{
    ret_code_t err_code = sd_app_evt_wait();
    APP_ERROR_CHECK(err_code);
}
//

static void wait_for_flash_ready(void)
{
    /* While fstorage is busy, sleep and wait for an event. */
    while (nrf_fstorage_is_busy(&fstorage))
    {
        power_manage();
    }
}
//

/**@brief Function for putting the chip into sleep mode.
 *
 * @note This function will not return.
 */
static void sleep_mode_enter(void)
{
    ret_code_t err_code;;
//    err_code = bsp_indication_set(BSP_INDICATE_IDLE);
//    APP_ERROR_CHECK(err_code);

    // Prepare wakeup buttons.
//    err_code = bsp_btn_ble_sleep_mode_prepare();
//    APP_ERROR_CHECK(err_code);

    // Go to system-off mode (this function will not return; wakeup will cause a reset).
    err_code = sd_power_system_off();
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for starting advertising.
 */
static void advertising_start(void)
{
    ret_code_t           err_code;
	
	err_code = ble_advertising_start(&m_advertising, BLE_ADV_MODE_FAST);
	APP_ERROR_CHECK(err_code);
    
    bsp_board_led_on(ADVERTISING_LED);
}
//

/**@brief Function for starting application timers. */
static void application_timers_start(void)
{
    ret_code_t err_code;

    // Start application timers.
    err_code = app_timer_start(m_battery_timer_id, BATTERY_LEVEL_MEAS_INTERVAL, NULL);
    APP_ERROR_CHECK(err_code);
}
//

/**@brief Function for application main entry.
 */
int main(void)
{
	uint32_t 	err_code;
	
    // Initialize.
    leds_init();
	
    // Initialize timers
    timers_init();
    
	// Initialize log
    log_init();
	
	// Initialize and load default values
	config_init();
	
	// Initialize buttons
    buttons_init();
    
    // Initialize ADC
    adc_configure();
	
	// Initialize ble stack
    ble_stack_init();
	
    // Initialize peer menager and security
	
	err_code = sd_ble_opt_set(BLE_GAP_OPT_PASSKEY, &ble_opt);
	APP_ERROR_CHECK(err_code);
	
    // Initialize gap
    gap_params_init();
	
    // Initialize gatt
    gatt_init();
	
    // Initialize services
    services_init();
	
    // Initialize advertisement
    advertising_init();
	
    // Initialize connection
    conn_params_init();

    // Initialize peer menager and sec
    peer_manager_init();

    // Start execution.
    application_timers_start();
	
    // Start execution.
    advertising_start();
    NRF_LOG_INFO("Blinky example started.");

    // Enter main loop.
    for (;;)
    {
		// any pending write request?
		if( pending_write == true )
		{
			// clear write request
			pending_write = false;
			
			// write to flash and check for errors
			err_code = nrf_fstorage_write( &fstorage, APP_FLASH_START_ADDRESS, &config, sizeof( config_t ), NULL );
			APP_ERROR_CHECK( err_code );
		}
		
        if (NRF_LOG_PROCESS() == false)
        {
            power_manage();
        }
    }
}


/**
 * @}
 */
